#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from pathlib import Path
import pickle
import math
import time
import multiprocessing
import itertools
from collections import defaultdict
import logging
import traceback
import re

import numpy as np
from gensim.models import FastText, Word2Vec

from .javalang.parse import parse
from .javalang.tree import *
from .javalang.tokenizer import Identifier
from .delimiter import Delimiter

MIN_IDENTIFIER = 5
MAX_IDENTIFIER = 200
MAX_DISTANCE = 20
SEQUENCE_LEN = 50
SEQ_PER_START = 2

MIN_WORD_COUNT = 10

IGNORE_NODES = (Annotation, )

def _build_identifier_graph(tree):
    ## solve member reference and method invocation
    identifiers = set(token.value for token in tree.tokens(Identifier))
    normalized_idens = set(Delimiter.split_camel(iden, to_lower=True).replace(" ", "_") for iden in identifiers)
    # print(normalized_idens)
    if not (MIN_IDENTIFIER <= len(normalized_idens) <= MAX_IDENTIFIER):
        return dict()
    leaf2ancestors = defaultdict(dict)
    for path, ast_node in tree:
        if isinstance(ast_node, IGNORE_NODES):
            continue
        path = (ast_node, ) + tuple(n for n in reversed(path) if isinstance(n, Node))
        path = path[:MAX_DISTANCE]
        for child in ast_node.children:
            if not (isinstance(child, str) and child in identifiers):
                continue
            leaf = Delimiter.split_camel(child, to_lower=True).replace(" ", "_")
            for dis, ancestor in enumerate(path, start=1):
                if ancestor in leaf2ancestors[leaf] and leaf2ancestors[leaf][ancestor] < dis:
                    continue
                leaf2ancestors[leaf][ancestor] = dis
            
    graph = dict()
    for leaf1, leaf2 in itertools.combinations(leaf2ancestors.keys(), 2):
        ancestors1, ancestors2 = leaf2ancestors[leaf1], leaf2ancestors[leaf2]
        common_ancestors = set(ancestors1.keys()) & set(ancestors2.keys())
        if len(common_ancestors) == 0:
            continue
        dis = min(ancestors1[common_anc] + ancestors2[common_anc] for common_anc in common_ancestors)
        if dis > MAX_DISTANCE:
            continue
        graph[leaf1, leaf2] = dis
        graph[leaf2, leaf1] = dis
    return graph

def _random_walk(graph):
    transition_table = defaultdict(list)
    for (u, v), dis in graph.items():
        transition_table[u].append((v, dis))
    # print(transition_table)
    for u, out_edges in list(transition_table.items()):
        neighbours, distances = zip(*out_edges)
        total_dis = sum(distances)
        weights = [math.log2(total_dis / dis) for dis in distances]
        total_weight = sum(weights)
        probs = [w / total_weight if total_weight > 0 else 1. for w in weights]
        transition_table[u] = neighbours, probs

    iden_seqs = list()   
    seq_len = min(len(transition_table), SEQUENCE_LEN) 
    for identifier in transition_table:
        for _ in range(SEQ_PER_START):
            seq = list()
            hop = 0
            cur_identifier = identifier
            while hop < seq_len:
                # seq.extend(slicer.slice(cur_identifier))
                seq.append(cur_identifier)
                neighbours, probs = transition_table[cur_identifier]
                cur_identifier = np.random.choice(neighbours, 1, probs)[0]
                hop += 1
            iden_seqs.append(seq)
    return iden_seqs

def _extract_for_subtree(tree):
    graph = _build_identifier_graph(tree)
    iden_seqs = _random_walk(graph)
    return iden_seqs

def _extract_for_code(code, level="method"):
    try:
        ast = parse(code)
    except:
        return set(), defaultdict(int)
    subtrees = [md for _, md in ast.filter((MethodDeclaration, ConstructorDeclaration))] if level == "method" else [ast]
    if len(subtrees) == 0:
        return set(), defaultdict(int)

    iden_seqs = list()
    word2count = defaultdict(int)
    for tree in subtrees:
        words = set()
        for seq in _extract_for_subtree(tree):
            # print(seq)
            delimited_seq = []
            for iden in seq:
                delimited_seq.append(iden)
                for word in iden.split("_"):
                    word2count[word] = word2count[word] + 1
            iden_seqs.append(delimited_seq)
        for word in words:
            word2count[word] = word2count[word] + 1      
    return iden_seqs, word2count

def generate_corpus(input_file, sequence_file, meta_file, level="method", buf_size=1000):
    with Path(input_file).open("rb") as f:
        codes = pickle.load(f)
    # print("finish loading code.")
    
    beg_time = time.time()
    seqs = list()
    word2count = defaultdict(int)
    seq_f = Path(sequence_file).open("w", encoding="utf-8")
    seq_num = 0
    num = 0
    for pid, cid, code in codes:
        num += 1
        try:
            _seqs, _word2count = _extract_for_code(code, level=level)
        except:
            logging.error(traceback.format_exc())
            continue
        seqs.extend(_seqs)
        for word, count in _word2count.items():
            word2count[word] = word2count[word] + count 
        if num % buf_size == 0 or num == len(codes):
            logging.info(f"processed code: {num}/{len(codes)}, vocab size: {len(word2count)}, cost time: {time.time() - beg_time}s.")
            seq_f.write("\n".join([" ".join(seq) for seq in seqs]))
            seq_f.flush()
            seq_num += len(seqs)
            seqs.clear()
        # if len(_seqs) == 0:
        #     print(code)
        # break
    seq_f.close()
    with Path(meta_file).open("wb") as f:
        pickle.dump((seq_num, word2count), f)
    

def train(corpus_files, meta_files, model_path=None, emb_path=None, model="fasttext", vector_size=100, window=15, min_count=MIN_WORD_COUNT, epochs=5, workers=32, fine_tuning=False):
    word2count = defaultdict(int)
    seq_num = 0
    for meta_file in meta_files:
        with Path(meta_file).open("rb") as f:
            _seq_num, _word2count = pickle.load(f)
        seq_num += _seq_num
        for word, count in _word2count.items():
            word2count[word] = word2count[word] + count
    logging.info(f"total examples: {seq_num}")

    class Corpus:
        def __init__(self, fnames):
            self.fnames = fnames
    
        def __iter__(self):
            for fname in self.fnames:
                with Path(fname).open("r", encoding="utf-8") as f:
                    sentences = f.readlines()
                    for sent in sentences:
                        yield re.split(r"[\s_]", sent.strip())
    
    model_cls = FastText if model == "fasttext" else Word2Vec
    if fine_tuning and model_path:
        model = model_cls.load(model_path)
    else:
        model = model_cls(sg=1, hs=1, vector_size=vector_size, window=window, min_count=min_count, workers=workers)
        model.build_vocab_from_freq(word2count)
    model.train(Corpus(corpus_files), total_examples=seq_num, epochs=epochs)
    if model_path:
        model.save(model_path)
    if emb_path:
        model.wv.save(emb_path)
    logging.info(f"vocab size: {len(model.wv.key_to_index)}")
    


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

from codetoolkit.javalang.parse import parse
from codetoolkit.javalang.tree import StatementExpression

if __name__ == "__main__":
    code = '''
    public class A {public static void wait ( JSONRPC self ) {
  @ ParamToProperty ( action = "idle" ) @ ParamToProperty ( action = "update" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty ( action = "timeout" ) @ ParamToProperty ( action = "package" ) @ ParamToProperty (}
    '''
    for _, node in parse(code).filter(StatementExpression):
        print(node)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-


from pathlib import Path
import os
from typing import List

from .candidate_provider import POSCandidateProvider
from .resource import DEFAULT_METHOD_CORPUS, DEFAULT_FIELD_CORPUS

class Node:
    def __init__(self, eles:List, tag:str):
        self.eles = eles
        self.tag = tag

    def __str__(self):
        return f"[{' '.join(str(ele) for ele in self.eles)}]:{self.tag}"

class Chunker:
    def __init__(self, provider=POSCandidateProvider(), word_freq:dict=None):
        self.provider = provider
        self.word_freq = word_freq
        if word_freq is None:
            self.__init_ferq_from_files(DEFAULT_METHOD_CORPUS, DEFAULT_FIELD_CORPUS)


    @classmethod
    def from_path(cls, method_path, field_path):
        instance = cls(POSCandidateProvider(), {})
        instance.__init_ferq_from_files(method_path, field_path)
        return instance

    
    def __init_ferq_from_files(self, method_path, field_path):
        word_freq = {}
        with Path(method_path).open("r", encoding="utf-8") as f:
            for line in f:
                words = line.strip().split()
                if words[0] not in word_freq:
                    word_freq[words[0]] = {}
                if words[-1] not in word_freq:
                    word_freq[words[-1]] = {}

                if len(words) == 1:
                    word_freq[words[0]]["method-single"] = word_freq[words[0]].get("method-single", 0) + 1
                else:
                    word_freq[words[0]]["method-begin"] = word_freq[words[0]].get("method-begin", 0) + 1
                    word_freq[words[-1]]["method-end"] = word_freq[words[-1]].get("method-end", 0) + 1
        with Path(field_path).open("r", encoding="utf-8") as f:
            for line in f:
                words = line.strip().split()
                if words[0] not in word_freq:
                    word_freq[words[0]] = {}
                if words[-1] not in word_freq:
                    word_freq[words[-1]] = {}

                if len(words) == 1:
                    word_freq[words[0]]["field-single"] = word_freq[words[0]].get("field-single", 0) + 1
                else:
                    word_freq[words[0]]["field-begin"] = word_freq[words[0]].get("field-begin", 0) + 1
                    word_freq[words[-1]]["field-end"] = word_freq[words[-1]].get("field-end", 0) + 1
        self.word_freq = word_freq

    def get_score(self, word):
        if word not in self.word_freq:
            return 0
        # print(self.word_freq[word])
        nverb = self.word_freq[word].get("method-begin", 0) + self.word_freq[word].get("field-single", 0)
        nnoun = self.word_freq[word].get("method-end", 0) + self.word_freq[word].get("field-single", 0) + self.word_freq[word].get("field-end", 0)
        if (nverb + nnoun) == 0:
            return 0
        return nverb / (nverb + nnoun)

    def chunk(self, phrase, is_construction):
        
        phrase = self.provider.provide(phrase)
        if is_construction:
            node = Node([" ".join(word for word, _ in phrase)], "NP")
            return Node([node], "ROOT")
        head, head_tags = phrase[0]
        if self.isSomeVerb(head_tags) and self.isSomeNoun(head_tags):
            score = self.get_score(head)
            # print(score)
            if score >= 0.8:
                phrase[0] = (head, ["baseV"])
            elif 0 < score <= 0.1:
                phrase[0] = (head, ["noun"])
        tail, tail_tags = phrase[-1]
        if self.isSomeVerb(tail_tags) and self.isSomeNoun(tail_tags):
            score = self.get_score(tail)
            if score >= 0.8:
                phrase[-1] = (tail, ["baseV"])
            elif 0 < score <= 0.2:
                phrase[-1] = (tail, ["noun"])

        rule_flag = False
        if not self.isSomeVerb(phrase[0][1]) and self.isSomeVerb(phrase[-1][1]):
            rule_flag = True

        if len(phrase) > 1 and "pp" in phrase[-1][1]:
            phrase[-2] = (phrase[-2][0], ["noun"])
        
        if len(phrase) > 1:
            first, first_tags = phrase[0]
            second, second_tags = phrase[1]
            first_score = self.get_score(first)
            second_score = self.get_score(second)
            if len(first_tags) == 1 and self.isSomeNoun(first_tags) and self.isSomeVerb(second_tags):
                if second_score > 0.3:
                    phrase[0] = (phrase[0][0], ["adv"])
            if len(second_tags) == 1 and (not self.isSomeVerb(first_tags)) and self.isSomeVerb(second_tags):
                if second_score > 0.3:
                    phrase[0] = (phrase[0][0], ["adv"])
            if self.isSomeVerb(first) and self.isSomeVerb(second):
                if second_score >= 0.8 and first_score <= 0.2:
                    phrase[0] = (phrase[0][0], ["adv"])
                elif second_score >= 0.99 and first_score <= 0.4:
                    phrase[0] = (phrase[0][0], ["adv"])

        output = self.NPVP(phrase[:])
        if rule_flag and output is not None:
            return output
        output = self.startsWithIsOrCan(phrase[:])
        if output is not None:
            return output
        output = self.VPPP(phrase[:])
        if output is not None:
            return output
        output = self.PP(phrase[:])
        if output is not None:
            return Node([output], "ROOT")
        output = self.VPNP(phrase[:])
        if output is not None:
            return output
        output = self.VPp(phrase[:])
        if output is not None:
            return output
        output = self.VPNM(phrase[:])
        if output is not None:
            return output
        output = self.VP(phrase[:])
        if output is not None:
            return Node([output], "ROOT")
        output = self.VPNPprepP(phrase[:])
        if output is not None:
            return output
        output = self.VPprepP(phrase[:])
        if output is not None:
            return output
        output = self.adjP(phrase[:])
        if output is not None:
            return Node([output], "ROOT")
        output = self.NP(phrase[:])
        if output is not None:
            return Node([output], "ROOT")
        output = self.NPVP(phrase[:])
        if output is not None:
            return output
        output = self.VPadjP(phrase[:])
        if output is not None:
            return output
        output = self.prepP(phrase[:])
        if output is not None:
            return Node([output], "ROOT")
        output = self.VPNPp(phrase[:])
        if output is not None:
            return output
        return None
        

    def VP(self, phrase):
        size = len(phrase)
        if size == 0:
            return None
        if size == 1 and self.isSomeVerb(phrase[0][1]):
            return Node([phrase[0][0]], "VP")
        last = phrase.pop()
        vm = self.VM(phrase[:])
        if size > 1 and vm is not None:
            return Node([vm, last[0]], "VP")
        return None

    def NP(self, phrase):
        if len(phrase) == 0:
            return None
        np_eles = []
        while len(phrase) > 0:
            if "prep" in phrase[0][1]:
                break
            np_eles.append(phrase.pop(0))
        if len(np_eles) == 0:
            return None
        np = None
        if self.isSomeNoun(np_eles[-1][1]) or "ingV" in np_eles[-1][1]:
            word, _ = np_eles.pop()
            if len(np_eles) == 0:
                np = Node([word], "NP")
            else:
                nm = self.NM(np_eles)
                if nm is not None:
                    np = Node([nm, word], "NP")
        if len(phrase) > 0 and np is not None:
            pP = self.prepP(phrase[:])
            if pP is not None and phrase[0][0] == "of":
                np.eles.append(pP)
            else:
                return None
        return np

    def PP(self, phrase):
        if len(phrase) == 0:
            return None
        if "pp" in phrase[-1][1]:
            _is = False
            if len(phrase) == 1:
                return Node([phrase[0][0]], "VP-PP")
            else:
                last = phrase.pop()
                if phrase[-1][0] == "is":
                    _is = True
                    phrase.pop()
                np = self.NP(phrase[:])
                if np is not None:
                    if _is:
                        return Node([np, "is", last[0]], "VP-PP")
                    else:
                        return Node([np, last[0]], "VP-PP")
                else:
                    return None
        return None

    def adjP(self, phrase):
        if len(phrase) == 0:
            return None
        if "adj" in phrase[-1][1]:
            if len(phrase) == 1:
                return Node([phrase[0][0]], "adjP")
            else:
                last = phrase.pop()
                np = self.NP(phrase[:])
                if np is not None:
                    return Node([np, last[0]], "adjP")
        return None

    def prepP(self, phrase):
        if len(phrase) <= 1:
            return None
        first = phrase.pop(0)
        out = self.NP(phrase[:]) 
        out = out if out is not None else self.PP(phrase[:])
        if "prep" in first[1] and out is not None:
            return Node([first[0], out], "prepP")
        return None

    def NM(self, phrase):
        if len(phrase) == 0:
            return None
        nm = []
        for word, tags in phrase:
            if self.isSomeNoun(tags) or self.isSomeVerb(tags) or "adj" in tags or "pp" in tags or "ingV" in tags or "quant" in tags or "pron" in tags or "prep" in tags:
                nm.append(word)
            else:
                return None
        return Node([" ".join(nm)], "NM")

    def VM(self, phrase):
        if len(phrase) == 0:
            return None
        vm = []
        for word, tags in phrase:
            if "adv" in tags:
               vm.append(word)
            else:
                return None
        return Node([" ".join(vm)], "VM")

    def VPNP(self, phrase):
        vp_eles = []
        while len(phrase) > 0:
            vp_eles.append(phrase.pop(0))
            if self.isSomeVerb(vp_eles[-1][1]):
                break
        vp = self.VP(vp_eles)
        np = self.NP(phrase[:])
        if vp is not None and np is not None:
            return Node([vp, np], "ROOT")
        return None

    def NPVP(self, phrase):
        vp_eles = []
        while len(phrase) > 0:
            vp_eles.insert(0, phrase.pop())
            if len(phrase) == 0 or self.isSomeNoun(phrase[-1][1]):
                break
        vp = self.VP(vp_eles)
        np = self.NP(phrase[:])
        if vp is not None and np is not None:
            return Node([np, vp], "ROOT")
        return None

    def VPadjP(self, phrase):
        vp_eles = []
        while len(phrase) > 0:
            vp_eles.append(phrase.pop(0))
            if self.isSomeVerb(vp_eles[-1][1]):
                break
        vp = self.VP(vp_eles)
        ap = self.adjP(phrase[:])
        if vp is not None and ap is not None:
            return Node([vp, ap], "ROOT")
        return None

    def VPNM(self, phrase):
        if len(phrase) == 0:
            return None
        if self.isSomeNoun(phrase[-1][1]):
            return None
        vp_eles = []

        while len(phrase) > 0:
            vp_eles.append(phrase.pop(0))
            if self.isSomeVerb(vp_eles[-1][1]):
                break
        vp = self.VP(vp_eles)
        nm = self.NM(phrase[:])
        if vp is not None and nm is not None:
            return Node([vp, nm], "ROOT")
        return None

    def VPPP(self, phrase):
        vp_eles = []
        while len(phrase) > 0:
            vp_eles.append(phrase.pop(0))
            if self.isSomeVerb(vp_eles[-1][1]):
                break
        vp = self.VP(vp_eles)
        pp = self.PP(phrase[:])
        if vp is not None and pp is not None:
            return Node([vp, pp], "ROOT")
        return None

    def VPprepP(self, phrase):
        vp_eles = []
        while len(phrase) > 0:
            vp_eles.append(phrase.pop(0))
            if self.isSomeVerb(vp_eles[-1][1]):
                break
        vp = self.VP(vp_eles)
        pp = self.prepP(phrase[:])
        if vp is not None and pp is not None:
            return Node([vp, pp], "ROOT")
        return None

    def VPNPprepP(self, phrase):
        vp_eles = []
        np_eles = []
        while len(phrase) > 0:
            vp_eles.append(phrase.pop(0))
            if self.isSomeVerb(vp_eles[-1][1]):
                break
        while len(phrase) > 0:
            if "prep" in phrase[0][1]:
                break
            np_eles.append(phrase.pop(0))
        vp = self.VP(vp_eles)
        np = self.NP(np_eles)
        pp = self.prepP(phrase[:])
        if vp is not None and np is not None and pp is not None:
            return Node([vp, np, pp], "ROOT")
        return None

    def VPp(self, phrase):
        last = phrase.pop()
        vp = self.VP(phrase[:])
        if vp is not None and "prep" in last[1]:
            return Node([vp, last[0]], "ROOT")

    def VPNPp(self, phrase):
        last = phrase.pop()
        vp_eles = []
        while len(phrase) > 0:
            vp_eles.append(phrase.pop(0))
            if self.isSomeVerb(vp_eles[-1][1]):
                break
        vp = self.VP(vp_eles)
        np = self.NP(phrase[:])
        if vp is not None and np is not None and "prep" in last[1]:
            return Node([vp, np, last[0]], "ROOT")
        return None

    def startsWithIsOrCan(self, phrase):
        if len(phrase) <= 1:
            return None
        if phrase[0][0] not in {"is", "can"}:
            return None
        first = phrase.pop(0)
        second = None
        if "ingV" in phrase[0][1] and first[0] == "is":
            if len(phrase) == 1:
                return Node([Node(["is", phrase[0][0]], "VP-ingV")], "ROOT")
            else:
                ing = phrase.pop(0)
                second = self.NP(phrase[:])
                if second is not None:
                    return Node([Node(["is", ing[0]], "VP-ingV"), second], "ROOT")
                second = self.prepP(phrase[:])
                if second is not None:
                    return Node([Node(["is", ing[0]], "VP-ingV"), second], "ROOT")
        if "baseV" in phrase[0][1] and first[0] == "can":
            if len(phrase) == 1:
                return Node([Node(["can", phrase[0][0]], "VP")], "ROOT")
            else:
                base = phrase.pop(0)
                second = self.NP(phrase[:])
                if second is not None:
                    return Node([Node(["can", base[0]], "VP"), second], "ROOT")
                second = self.prepP(phrase[:])
                if second is not None:
                    return Node([Node(["can", base[0]], "VP"), second], "ROOT")
        second = self.PP(phrase[:])
        if second is not None:
            return Node([Node([first[0]], "VP"), second], "ROOT")
        second = self.adjP(phrase[:])
        if second is not None:
            return Node([Node([first[0]], "VP"), second], "ROOT")
        second = self.NP(phrase[:])
        if second is not None:
            return Node([Node([first[0]], "VP"), second], "ROOT")
        second = self.NM(phrase[:])
        if second is not None:
            return Node([Node([first[0]], "VP"), second], "ROOT")
        return None
            
    def isSomeVerb(self, tags):
        return "baseV" in tags or "irV" in tags or "3PS" in tags
        
    def isSomeNoun(self, tags):
        return "noun" in tags or "plN" in tags or "quant" in tags or "pron" in tags


#!/usr/bin/env python
# -*- encoding: utf-8 -*-


from nltk.corpus import wordnet as wn
import re

from .resource import DEFAULT_DICT

class POSCandidateProvider:
    def __init__(self, dictionary=DEFAULT_DICT):
        self.dictionary = dictionary
        self.pos_cache = {}

    def provide(self, phrase):
        words = phrase.split()
        tags = []
        for word in words:
            candidates = []
            if self.isArtical(word):
                tags.append((word, ["art"]))
                continue
            if self.isPreposition(word):
                tags.append((word, ["prep"]))
                continue
            if self.isQuantifier(word):
                tags.append((word, ["quant"]))
                continue
            if self.isIrregularV(word):
                candidates.append("irV")
            if self.isAdverb(word):
                candidates.append("adv")
            if self.isPronoun(word):
                candidates.append("pron")
            if self.isAdjective(word):
                candidates.append("adj")
            if self.isNoun(word):
                candidates.append("noun")
            if self.isPlNoun(word):
                # if len(candidates) > 0 and candidates[-1] == "noun":
                #     candidates.pop()
                candidates.append("plN")
            if self.is3PS(word):
                candidates.append("3PS")
            if self.isBaseVerb(word):
                # if not (len(candidates) > 0 and candidates[-1] == "3PS"):    
                candidates.append("baseV")
            if self.isIngVerb(word):
                # if len(candidates) > 0 and candidates[-1] == "baseV":
                #     candidates.pop()
                candidates.append("ingV")
            if self.isPastVerb(word):
                # if len(candidates) > 0 and candidates[-1] == "baseV":
                #     candidates.pop()
                candidates.append("pastV")
            if self.isPP(word):
                candidates.append("pp")
            if len(candidates) == 0:
                candidates.append("noun")

            tags.append((word, candidates))
        return tags

    def pos_in_wn(self, word):
        lemma = word.lower()
        if lemma in self.pos_cache:
            return self.pos_cache[lemma]

        get_synset = wn.synset_from_pos_and_offset
        index = wn._lemma_pos_offset_map
        synsets = [
            get_synset(p, offset)
            for p in [wn.NOUN, wn.VERB, wn.ADJ, wn.ADV]
            for offset in index[lemma].get(p, [])
        ]
        self.pos_cache[lemma] = {s.pos() for s in synsets}
        return self.pos_cache[lemma]
            

    def isArtical(self, word):
        if word in {"a", "an", "the"}:
            return True
        return  False

    def isPreposition(self, word):
        if word in self.dictionary["PREP"]:
            return True
        return False

    def isQuantifier(self, word):
        if word in self.dictionary["QUAN"]:
            return True
        return False

    def isIrregularV(self, word):
        if word in self.dictionary["IRRV"]:
            return True
        return False

    def isAdverb(self, word:str):
        if word in self.dictionary["ADV"]:
            return True
        wn_pos = self.pos_in_wn(word)
        if wn.ADV in wn_pos:
            self.dictionary["ADV"].add(word)
            
            return True
        if len(wn_pos) > 0 and len(wn_pos & {wn.VERB, wn.NOUN, wn.ADJ, wn.ADV, wn.ADJ_SAT}) > 0:
            return False
        if word.endswith("ly"):
            base = word[:-2]
            if len(base) > 0 and base[-1] == "i":
                base = base[:-1] + "y"
            if base in self.dictionary["VOCAB"]:
                self.dictionary["ADV"].add(word)
                
                return True
        return False

    def isPronoun(self, word):
        if word in self.dictionary["PRON"]:
            return True
        return False

    def isAdjective(self, word:str):
        if word in self.dictionary["ADJ"]:
            return True
        wn_pos = self.pos_in_wn(word)
        if wn.ADJ in wn_pos or wn.ADJ_SAT in wn_pos:
            self.dictionary["ADJ"].add(word)
            
            return True
        if len(wn_pos) > 0 and len(wn_pos & {wn.VERB, wn.NOUN, wn.ADJ, wn.ADV, wn.ADJ_SAT}) > 0:
            return False
        if word.endswith("able") and len(word) > 6:
            self.dictionary["ADJ"].add(word)
            
            return True
        if word.endswith("er") or word.endswith("est"):
            if word.endswith("er"):
                base = word[:-2]
            else:
                base = word[:-3]
            
            if len(base) > 0 and base[-1] == "i":
                base = base[:-1] + "y"
            if len(base) >= 2 and base[-1] == base[-2]:
                base = base[:-1]
            if self.isAdjective(base):
                self.dictionary["ADJ"].add(word)
                
                return True
        prefix = word
        if len(word) > 1 and word.endswith("y") and word[-2] not in {"a", "e", "i", "o", "u"}:
            prefix = word[:-1] + "i"
        if prefix + "ly" in self.dictionary["VOCAB"] or prefix + "ness" in self.dictionary["VOCAB"]:
            self.dictionary["ADJ"].add(word)
            
            return True

        return False

    def isNoun(self, word:str):
        if word in self.dictionary["NOUN"]:
            return True
        wn_pos = self.pos_in_wn(word)
        if wn.NOUN in wn_pos:
            return True
        if len(wn_pos) > 0 and len(wn_pos & {wn.VERB, wn.NOUN, wn.ADJ, wn.ADV, wn.ADJ_SAT}) > 0:
            return False
        if word + "ence" in self.dictionary["VOCAB"] or  word + "ance" in self.dictionary["VOCAB"]:
            return False
        if word in self.dictionary["NABBR"]:
            return True
        elif self.isIrregularV(word):
            return False
        if re.match(r"^.+(ity|tion|is[tm]|ness|or)$", word):
            self.dictionary["NOUN"].add(word)
            
            return True

        prefix = re.sub(r"([^aeiou])y$", r"\1i", word)
        if re.match(r"^.*(i|s|ch|x|z|sh)$", prefix) and prefix + "es" in self.dictionary["VOCAB"]:
            self.dictionary["NOUN"].add(word)
            return True

        if word + "s" in self.dictionary["VOCAB"]:
            self.dictionary["NOUN"].add(word)
            
            return True

        return False

    def isPlNoun(self, word:str):
        base = word
        if word.endswith("es"):
            base = word[:-2]
            if len(base) > 1 and base.endswith("i") and base[-2] not in {"a", "e", "i", "o", "u"}:
                base = base[:-1] + "y"
            if self.isNoun(base) and re.match(r"^.*(y|s|ch|x|z|sh)$", base):
                return True
        if re.match(r"(.*[^sui])s$", word) and len(word) > 3 and wn.NOUN in self.pos_in_wn(word[:-1]):
            return True
        if re.match(r"(.*[^sui])s$", word) and len(word) > 4 and self.isNoun(word[:-1]):
            return True
        return False

    def is3PS(self, word:str):
        if re.match(r"(.*[^sui])s$", word):
            base = word[:-1]
            base = re.sub(r"ie$", "y", base)
            if self.isBaseVerb(base):
                return True
        return False

    def isBaseVerb(self, word):
        origin = word
        if word in self.dictionary["VERB"]:
            return True
        if self.isIrregularV(word):
            return False
        if wn.VERB in self.pos_in_wn(word):
            return True
        if len(self.pos_in_wn(word)) > 0 and len(self.pos_in_wn(word) & {wn.VERB, wn.NOUN, wn.ADJ, wn.ADV, wn.ADJ_SAT}) > 0:
            return False

        if word not in self.dictionary["VOCAB"] or len(word) < 2:
            return False

        if re.match(r"^(re|en).*", word) and self.isBaseVerb(word[2:]):
            return True
        if re.match(r"(.*)(ize|ify)$", word):
            return True

        double = word
        if re.match(r"[aeiou][ngdtlp]$", word):
            double += double[-1]
        elif re.match(r".*([^ey])e$", word):
            word = word[:-1]
        elif re.match(r"^.*y$", word):
            word = word[:-1] + "i"
        
        if word + "s" in self.dictionary["VOCAB"] or word + "es" in self.dictionary["VOCAB"] or double + "es" in self.dictionary["VOCAB"]:
            if word + "ing" in self.dictionary["VOCAB"] or double + "ing" in self.dictionary["VOCAB"]:
                self.dictionary["VERB"].add(origin)
                self.dictionary["VOCAB"].add(origin)
                return True
            if word + "ed" in self.dictionary["VOCAB"] or double + "ed" in self.dictionary["VOCAB"]:
                self.dictionary["VERB"].add(origin)
                self.dictionary["VOCAB"].add(origin)
                return True
        return False

    def isIngVerb(self, word:str):
        if word.endswith("ing") and self.isBaseVerb(word[:-3]):
            return True
        return False

    def isPastVerb(self, word:str):
        if word.endswith("ed") and self.isBaseVerb(word[:-2]):
            return True
        return False

    def isPP(self, word:str):
        if word in self.dictionary["PP"]:
            return True
        elif re.match(r"(.*)e[nd]$", word):
            base = word[:-2]
            base = re.sub(r"i$", "y", base)
            if self.isBaseVerb(base) or self.isBaseVerb(base + "e") or self.isBaseVerb(base[:-1]):
                return True
        return False




'''Ronin: an identifier splitter based on the Samurai algorithm

Introduction
------------

Spiral is a Python 3 package that implements numerous identifier splitting
algorithms.  Identifier splitting is the task of breaking apart program
identifier strings such as 'getInt' or 'readUTF8stream' into component
tokens: ['get', 'int'] and ['read', 'utf8', 'stream'].  The need for
splitting identifiers arises in a variety of contexts, including natural
language processing (NLP) methods applied to source code analysis and program
comprehension.

Ronin is a Python module that implements an approach that was originally
based on Samurai, described in the following paper:

  Enslen, E., Hill, E., Pollock, L., & Vijay-Shanker, K. (2009).  Mining
  source code to automatically split identifiers for software analysis.  In
  Proceedings of the 6th IEEE International Working Conference on Mining
  Software Repositories (MSR'09) (pp. 71-80).

The implementation of Ronin in the present Python module was created by the
author from the description of the algorithm published in the paper, then
modified repeatedly in an attempt to improve performance.  A significant goal
in Ronin was to produce a splitter that had acceptable performance without
having to use a local frequency table derived from a source code base being
processed.  Ronin uses only a global frequency table derived from mining tens
of thousands of GitHub project repositories containing Python source code
files.  It also uses a dictionary of English words (from NLTK) to help
produce more "natural" splits.  Ronin has many notable changes compared to
Samurai and many tunable parameters.

Usage
-----

Note #1: for fastest performance, using the optimization options provided by
the Python interpreter (i.e., the options -O or -OO).

Note #2: If you do not call init() separately, then the first time you call
ronin.split(...) it will call init() itself.  This will make the first call
to split(...)  take much longer to run than subsequent invocation.  Please
rest assured that this is only a one-time startup cost for a given session.

The simplest way to use this module in a Python program is to import the
module from the Spiral package and then call ronin.split(...) with a
string.  For example,

    from spiral import ronin
    result = ronin.split('someidentifier')

The function split(...) produces a list of strings as its output.  It will do
its best-guess at how a given identifier should be split using a built-in
table of term frequencies, prefixes, and suffixes.  The built-in table of
term frequencies came from an analysis of over 46,000 randomly selected
software projects in GitHub that contained at least one Python source code
file.

Ronin's performance can be tuned by providing it with a different table of
term frequencies.  If you want to experiment with using a different set of
global frequencies, you can supply them to the function init(...) once,
before making any calls to split(...).  (This is done so that the cost of
computations involved in the frequencies is only done once.)  Use the keyword
argument 'frequencies':

    from spiral import ronin
    ronin.init(frequencies = myfrequencies)
    result = ronin.split('someidentifier')

The format of the value given to 'frequencies' should be a dictionary where
the keys are the terms and the values are integers representing the number of
times the term appears in the code base.  A term that appears once will have
a value of 1, a term that appears twice will have a value of 2, etc.  Ronin
will normalize the values.  (Do NOT normalize the values yourself, as this
will confuse the algorithm.)  Note that generating the global table is not a
trivial undertaking.  It requires mining a large number of source code
repositories to extract identifiers of classes, functions, variables, etc.,
then splitting them to produce individual identifier fragments, and finally
curating the result manually.

Finally, init(...) accepts additional arguments for tuning some other
parameters that control how it calculates costs of splitting identifiers.
The default values of the parameters were determined using numerical
optimization methods.  Please see the documentation for init(...) for more
information.

Tracing the algorithm
---------------------

To print what the splitter is doing while processing a given string, you can
turn on logging by setting the logging level to "logging.DEBUG" before
importing the Ronin module.  Here is an example:

  import logging
  logging.basicConfig(level = logging.DEBUG, format = '')
  from spiral import ronin

Logging will be printed to the standard output stream.  Note: this will only
work if you do NOT use the -O option to the Python interpreter.

Optimization
------------

This module wraps all logging code with the Python __debug__ compile-time
symbol, so that you can use the -O flag to Python to make it omit the code
completely.  This is more efficient than simply setting the logging level.

Explanation of the name
-----------------------

The name "Ronin" is a play on the use of the name "Samurai" by Enslen et al.
(2009) for their identifier splitting algorithm.  The bulk of Ronin is
identical to Samurai, but it would be inappropriate to call this
implementation Samurai too.  In an effort to imply the lineage of this
modified algorithm, I chose "Ronin" (a name referring to a drifter samurai
without a master, during the Japanese feudal period).

Authors
-------

Michael Hucka <mhucka@caltech.edu>

Copyright
---------

Copyright (c) 2017 by the California Institute of Technology.  This software
was developed as part of the CASICS project, the Comprehensive and Automated
Software Inventory Creation System.  For more, visit http://casics.org.
'''

from   bisect import insort
import math
from   nltk.stem import SnowballStemmer
import os
import re
import sys

try:
    sys.path.append(os.path.join(os.path.dirname(__file__), ".."))
except:
    sys.path.append("..")

try:
    from .frequencies import *
    from .simple_splitters import heuristic_split, elementary_split
    from .constants import *
except:
    from frequencies import *
    from simple_splitters import heuristic_split, elementary_split
    from constants import *


# Global constants.
# .............................................................................

# If the caller doesn't supply a frequency table, we use a default one.
# The file needs to be kept in the same directory as this Python module file.

try:
    _DATA_DIR = os.path.join(os.path.dirname(__file__), 'data')
except:
    _DATA_DIR = os.path.dirname('./data')

_DEFAULT_FREQ_PICKLE = os.path.join(_DATA_DIR, 'frequencies.pklz')
'''Pickle file storing the default global frequency table shipped with this
module.  This constant is only read by ronin.init(...).'''

_DICTIONARY_PICKLE = os.path.join(_DATA_DIR, 'dictionary.pklz')
'''Pickle file storing the dictionary used by Ronin.  See the code in
init() for the exact contents, but to summarize, it's comprised of some NLTK
dictionaries and the special_computing_terms set from constants.py'''


# Main class
# .............................................................................

class Ronin(object):
    _dictionary         = None
    _stemmer            = None
    _frequencies        = None
    _highest_freq       = None
    _length_cutoff      = None
    _short_min_freq     = None
    _low_freq_cutoff    = None
    _normal_exponent    = None
    _dict_word_exponent = None
    _camel_bias         = None
    _recognition_bias   = 0.0000003635
    _biased_threshold   = 0
    _alt_exponent       = 1
    _exact_case         = False

    # The default parameter values came from optimizing against the INTT data
    # set/oracle, then doing some hand-tweaking of the parameter values.
    # The final score for INTT was:
    #    INTT:   17287/18772 (92.09%)
    # Using the Ludiso oracle/data set as a test set, with the same parameter
    # values, the following is the accuracy:
    #    Ludiso: 2248/2663   (84.42%)
    #
    # Many of the "failures" against these sets of identifiers are actually
    # not failures, but cases that Ronin clearly gets a more correct answer
    # or where there is a legitimate difference in interpretation.  Examples:
    #
    #   Token            Ludiso               Ronin
    #   --------         ----------------     -------------
    #   fread            fread                f, read
    #   a.ecart_         a, ecart             a, e, cart
    #   ReadUtf8Z        Read, Utf, 8, Z      Read, Utf8, Z
    #   MAPI_BCC         MAPI, BCC            M, API, BCC
    #   GetWSIsEnabled   Get, WSIs, Enabled   Get, WS, Is, Enabled
    #
    # Names like 'fread' could be considered appropriate to leave alone, but
    # 'fread' actually stands for 'file read' and thus IMHO it makes sense to
    # split it -- splitting identifiers into meaningful tokens is the purpose
    # of Ronin, after all.
    #
    # Some other differences with Ludiso are due to Ronin splitting terms
    # that are more typically considered separate words but are treated as
    # one word in Ludiso, or vice versa.  Examples:
    #
    #   Token                Ludiso                  Ronin
    #   --------             ----------------        -------------
    #   FF_LOSS_COLORSPACE   FF, LOSS, COLORSPACE    FF, LOSS, COLOR, SPACE
    #   m_bFilenameMode      m, b, File, name, Mode  m, b, Filename, Mode
    #
    # (The main entry in Wikipedia for "color space" is two words, while for
    # "filename" it is indeed one word.)  So, Ronin's performance may
    # actually be better than the pure numbers imply.  However, without going
    # through every Ludiso case and manually reinterpreting the splits, we
    # can't say for sure whether it really is.
    #
    # All that aside, it is definitely the case that Ronin gets many cases
    # wrong.  It is not perfect by any means.
    #
    # Other notes:
    #
    # * Ronin is tuned for splitting program identifiers, which is NOT the
    #   same as splitting strings of concatenated words.  With its default
    #   parameter values, it is not good at splitting strings like
    #   "driveourtrucks" which are not composed of tokens commonly
    #   encountered in source code contexts.
    #
    # * Ronin was trained on source code repositories containing Python code.
    #   The default frequency table may not be optimal for splitting things
    #   that come from non-Python source files, although the test results on
    #   Ludiso and INTT show that it is pretty good on non-Python content.
    #
    # * Reducing the value of parameter low_freq_cutoff increases Ronin's
    #   tendency to split tokens, but at the expense of worsening scores on
    #   the INTT and Ludiso data sets.  However, it's notable that reducing
    #   the value to something like 10 makes Ronin more likely to split
    #   concatenated word strings like "driveourtrucks" or "societynamebank".
    #
    def init(self, frequencies = None, exact_case = False,
             low_freq_cutoff = 340, length_cutoff = 2, short_min_freq = 286540,
             normal_exponent = 0.1527, dict_word_exponent = 0.1228,
             camel_bias = 8.6319, recognition_bias = 0.0000003635,
             alt_exponent = 1.2165):
        '''Initialize internal frequency files for the Ronin split() function.
        Note: the first time this function is called, it will take noticeable
        time because it will load a global frequency table (unless one is
        provided to override the default) and some NLTK dictionaries.

        Parameter 'frequencies' should be a dictionary where the keys are the
        terms and the values are integers representing the number of times
        the term appears in the code base.  A term that appears once will
        have a value of 1, a term that appears twice will have a value of 2,
        etc.  If not provided, Ronin will use a default table.  (Note that
        generating the global table is not a trivial undertaking.  It
        requires mining a large number of source code repositories to extract
        identifiers of classes, functions, variables, etc., then splitting
        them to produce individual identifier fragments, and finally curating
        the result manually.)

        Parameter 'exact_case' indicates whether the identifiers in the
        frequency table should be kept as-is, or whether only the lower-case
        form of the identifiers should be kept.  If False (the default),
        different case variations will be discarded and the maximum frequency
        of any case variation of a given identifier will be kept.  (E.g., if
        both 'foo' with frequency 20 and 'FOO' with frequency 1000 are
        present, the final table will have an entry for 'foo' with value
        1000.)  If True, all the case variants in the frequency table will be
        kept, and during scoring, the following alternatives will be tried in
        this order when searching for tokens in the table: try exact case
        match first, then capitalizing the token ('fOoO' -> 'Fooo'), and
        finally lower-casing the token ('Foo' or 'FOO' -> 'foo').  The first
        match will be returned.

        This init function also accepts a number of optional arguments that
        control internal parameter values.  The default values were
        determined by numerical optimization in conjunction with the default
        global frequency table shipped with Ronin.  In case you want to try
        searching for other values for a particular application, init()
        offers the following adjustable parameters, but beware that the
        values for optimal performance depend very much on the specific
        global frequency table you use and finding good values requires
        optimizing all the parameters simultaneously -- a compute-intensive
        process:

         * 'low_freq_cutoff': used when scoring strings lacking camel case
           transitions, this is a cut-off value below which a given frequency
           value in the frequency table is treated as being 0.  This needs to
           have a value greater than 0 to have any effect.  For example, if
           the cutoff is set to 10, any frequency less than or equal to 10
           will be given a score of 0.  This threshold tends to counteract
           noisiness in global frequency tables created from a large number
           of disparate sofware projects.  (Note: the default frequency table
           provided with Ronin does not have entries with values less than 10
           in order to keep the file size small, but a full table is also
           provided with Spiral.  See the 'spiral/data' subdirectory of the
           package.)

         * 'length_cutoff': sets a lower limit for string lengths, such that
           the frequency for any string shorter or equal to 'length_cutoff'
           needs to be higher than 'short_min_freq' or else the
           frequency value is taken to be 0 instead.

         * 'short_min_freq': minimum frequency of a short string that
           will be accepted.  If a string in the frequency table is shorter
           than 'length_cutoff' and its frequency value is lower than this,
           then its value will be taken as 0 instead.

         * 'normal_exponent': the value of an exponent in the formula used
           to adjust the frequency scores before they are used in the splitter
           algorithm.

         * 'dict_word_exponent': the value of an exponent in the formula used
           to adjust the frequency scores before they are used in the splitter
           algorithm, for the case when the token is a dictionary word.

         * 'camel_bias': multiplier used in a score comparison when deciding
           where to split string segments with an upper-to-lower case
           transition.

         * 'recognition_bias': multiplier used in a score comparison when a
           token is recognized (e.g., due to being a dictionary word or being
           defined in a list of common special terms).  This factor is
           multiplied by the highest frequency in the frequency table, so if
           the highest frequency is very large, the value of recognition_bias
           needs to be extremely small to have any effect.

         * 'alt_exponent': an exponent used in rescaling scores when there is
           more than one possible split to choose from.

        Changing only a single parameter value is unlikely to yield good
        results.  Beware of testing changes on a few chosen examples: you may
        be able to get acceptable results for those few cases but will likely
        find that performance worsens on other cases.  Adjusting parameter
        values is really best done using multiparameter optimization methods.
        An optimization utility is included in the Spiral source code
        distribution, to help find parameter values for the above.
        '''
        if not self._frequencies:
            if not frequencies:
                if os.path.exists(_DEFAULT_FREQ_PICKLE):
                    frequencies = frequencies_from_pickle(_DEFAULT_FREQ_PICKLE)
                else:
                    raise ValueError('Cannot read frequencies pickle file "{}"'
                                     .format(_DEFAULT_FREQ_PICKLE))
            if exact_case:
                self._frequencies = frequencies
            else:
                newdict = {}
                for key, value in frequencies.items():
                    lc_key = key.lower()
                    if lc_key in newdict:
                        newdict[lc_key] = max(value, newdict[lc_key])
                    else:
                        newdict[lc_key] = value
                self._frequencies = newdict
        self._exact_case = exact_case
        self._highest_freq = max(self._frequencies.values())
        self._recognition_bias = recognition_bias
        self._biased_threshold = self._recognition_bias * self._highest_freq
        self._low_freq_cutoff = low_freq_cutoff
        self._length_cutoff = length_cutoff
        self._short_min_freq = short_min_freq
        self._normal_exponent = normal_exponent
        self._dict_word_exponent = dict_word_exponent
        self._camel_bias = camel_bias
        self._alt_exponent = alt_exponent
        if not self._dictionary:
            if os.path.exists(_DICTIONARY_PICKLE):
                import gzip, pickle
                with gzip.open(_DICTIONARY_PICKLE, 'rb') as pickle_file:
                    self._dictionary = pickle.load(pickle_file)
            else:
                from nltk.corpus import words as nltk_words
                from nltk.corpus import wordnet as nltk_wordnet
                # Note: I also tried adding the words from /usr/share/dict/web,
                # but the only additional words it had that were not already in
                # the next two dicts were people's proper names. Not useful.
                self._dictionary = set(nltk_words.words())
                self._dictionary.update(nltk_wordnet.all_lemma_names())
                self._dictionary.update(special_computing_terms)
        self._stemmer = SnowballStemmer('english')
        # Generate scoring function based on exact case flag.  Do it here so
        # we don't have to keep testing the variable at run-time.
        if exact_case:
            def score_function(token):
                lc_token = token.lower()
                if lc_token in common_terms_with_numbers:
                    return self._highest_freq
                if token in self._frequencies:
                    return self._frequencies[token]
                cap_token = token.capitalize()
                if cap_token in self._frequencies:
                    return self._frequencies[cap_token]
                if lc_token in self._frequencies:
                    return self._frequencies[lc_token]
                return 0
        else:
            def score_function(token):
                lc_token = token.lower()
                if lc_token in common_terms_with_numbers:
                    return self._highest_freq
                if lc_token in self._frequencies:
                    return self._frequencies[lc_token]
                return 0
        self._score = score_function


    def split(self, identifier, keep_numbers=True):
        '''Split the given identifier string at token boundaries using a
        combination of heuristics and token frequency tables.  This function
        produces a list of strings as its output.

        Parameter 'keep_numbers' determines the handling of numbers.  When
        True, all digits are kept.  When False, the behavior is as follows:
        leading digits are removed from the identifier string but other
        digits are retained, then hard + camel-case splitting is performed,
        then (1) embedded digits are removed if they appear at the tail ends
        of tokens in the split, but not if they appear in the middle of
        tokens or are recognized as being part of common abbreviations such
        as 'utf8'; and (2) tokens that consist ONLY of digits are removed.

        If you have a set of frequencies for your project, you can supply
        them to the function init(...) once, before making calls to split().
        This function will otherwise use a built-in global table of
        frequencies.

        Note: calling init(...) is optional, but if you do not call it, then
        this function will call it the first time it is invoked.  Consequently,
        the first time split(...) is called, you may experience a noticeable
        delay while it calls init(...) to load various data files.  Subsequent
        invocations of split(...) will not be as slow.

        The Ronin algorithm is a significant modification of the Samurai
        algorithm described by Eric Enslen, Emily Hill, Lori Pollock and
        K. Vijay-Shanker in a 2009 paper titled "Mining source code to
        automatically split identifiers for software analysis" (Proceedings
        of the 6th IEEE International Working Conference on Mining Software
        Repositories, 2009).
        '''
        splits = []
        camel = False
        if not self._frequencies:
            self.init()
        initial_list = heuristic_split(identifier, keep_numbers=keep_numbers)
        for s in initial_list:
            if self._recognized(s):
                splits = splits + [s]
                continue
            # Look for upper-to-lower case transitions.
            transition = re.search(r'[A-Z][a-z]', s)
            if not transition:
                parts = [s]
            else:
                i = transition.start(0)
                camel = s[i:] if i > 0 else s
                camel_score = self._score(camel)
                alt = s[i+1:]
                # Logically, this next comparison should use the raw_score
                # rather than the adjusted or rescaled score.  Yet doing that
                # leads to worse performance even after optimizing the
                # parameters, and I don't know why.  Using the adjusted score
                # here seems like an apples-to-oranges comparison because of
                # the heavy rescaling applied by _adjusted_score().  Still,
                # the optimization results are clear and consistent.
                alt_score = self._adjusted_score(alt) * self._camel_bias
                if camel_score >= alt_score:
                    parts = [s[0:i], s[i:]] if i > 0 else [s]
                else:
                    parts = [s[0:i+1], s[i+1:]]
            splits = splits + parts

        results = []
        for token in splits:
            score = self._adjusted_score(token)
            results = results + self._same_case_split(token, score)[1]
        return results


    def _same_case_split(self, s, score_ns = .0000005, level = 0):
        # Returns a tuple of (max_score, split).
        # 'level' is to detect recursive calls, and also indent debug output.
        if self._recognized(s):
            return (score_ns, [s])
        new_split = None
        max_index = len(s) - 1
        max_score = -1
        threshold = max(self._adjusted_score(s), score_ns)
        splits = []
        i = 0
        while i < max_index:
            i += 1
            left, right = s[0:i], s[i:]
            prefix = self._is_prefix(left) or (len(s) > 5 and self._is_suffix(right))

            if prefix:
                continue

            score_l = self._adjusted_score(left)
            score_r = self._adjusted_score(right)
            break_r = score_r > threshold
            break_l = score_l > threshold
            either_score = max(score_l, score_r)

            if break_l and break_r:
                if (score_l + score_r) > max_score:
                    max_score = score_l + score_r
                    new_split = [left, right]
                    insort(splits, (max_score, new_split))
            elif break_l:
                if (either_score > self._biased_threshold and self._recognized(right)):
                    insort(splits, (either_score, [left, right]))
                else:
                    ts, tmp = self._same_case_split(right, score_ns, level + 1)
                    if tmp[0] != right:
                        new_split = [left] + tmp
                        insort(splits, (ts/len(new_split), new_split))
                    elif self._special_case(right):
                        max_score = either_score
                        insort(splits, (either_score, [left, right]))
            elif break_r and (self._recognized(left)
                              or len(left) <= self._length_cutoff
                              or self._special_case(right)):
                if either_score > self._biased_threshold:
                    insort(splits, (score_r/len(left), [left, right]))
                elif self._recognized(right):
                    insort(splits, (score_r, [left, right]))

        if splits:
            alt_score, alt_split = splits[-1]
            scaled_score = alt_score/math.pow(len(alt_split), self._alt_exponent)
            if scaled_score >= score_ns:
                result = splits[-1]
            else:
                result = (score_ns, [s])
        else:
            result = (max_score, [s])
        return result


    def _rescaled_score(self, token):
        '''Return the rescaled score for the given 'token', without applying
        the thresholding step that _adjusted_score() applies.'''
        return self._rescale(token, self._score(token))


    def _adjusted_score(self, token):
        '''Return the thresholded, adjusted score for the given 'token'.'''
        length = len(token)
        if length == 0:
            return 0
        score_value = self._score(token)
        if length <= self._length_cutoff and not self._special_case(token):
            if score_value <= self._short_min_freq:
                return 0
        if score_value <= self._low_freq_cutoff:
            return 0
        return self._rescale(token, score_value)


    def _rescale(self, token, score_value):
        '''Rescale the given token score value.'''
        if self._recognized(token):
            return math.pow(score_value, self._dict_word_exponent)
        else:
            return math.pow(score_value, self._normal_exponent)


    def _recognized(self, token):
        tlower = token.lower()
        return self._special_case(tlower) or self._in_dictionary(tlower)


    def _special_case(self, token):
        return (token in common_terms_with_numbers
                or token in special_computing_terms
                or self._stem(token) in special_computing_terms)


    def _in_dictionary(self, word):
        if len(word) <= 1:
            # Counting single-letter words turns out to be unhelpful.
            return False
        return (word in self._dictionary or self._stem(word) in self._dictionary)


    def _stem(self, token):
        # The NLTK stemmer sometimes gives weird results, and the particular
        # case of weird technical terms ending in 's' has been most troublesome.
        if len(token) > 1 and token[-1:] == 's':
            return token[:-1]
        else:
            return self._stemmer.stem(token)


    def _is_prefix(self, s):
        return s.lower() in prefixes


    def _is_suffix(self, s):
        return s.lower() in suffixes


# Module entry points.
# .............................................................................

_RONIN_INSTANCE = Ronin()

# The following are the functions we tell people to use:

init  = _RONIN_INSTANCE.init
split = _RONIN_INSTANCE.split



'''
constants: constants used in Spiral splitters.
'''

common_suffix_numbers = {'16', '32', '64', '128', '256', '512', '1024'}
'''
List of numbers that are commonly put after some other strings, to form
symbols such as "int32", "float64", etc.
'''

# General principles for the following:
# 1. Only put in terms that at least one digit in them.
# 2. Avoid 2-character sequences, because they match too many things and
#    cause too many bad splits.
# 3. Stick to computing terms that many people use or have used, not stuff
#    that a particular program happens to use.
# 4. Generally, avoid packages for particular languages, like "urllib3".
#
# Sources for the following:
# - https://en.wikipedia.org/wiki/List_of_computing_and_IT_abbreviations
# - https://en.wikipedia.org/wiki/List_of_codecs
# - https://en.wikipedia.org/wiki/Comparison_of_video_codecs
# - the book 'The Definitivei Guide to iReport' by Toffoli, 2007
# - various lists and dictionaries of computing terms found by Googling
# - personal experience
 
common_terms_with_numbers = {
    '1st',
    '1to1',
    '2fa',
    '2nd',
    '2of7',
    '3d',
    '3of9',
    '3ivx',
    '3rd',
    '88k',
    'a52',
    'ac2',
    'ac3',
    'ada83',
    'ada95',
    'amd64',
    'asn1',
    'avs1',
    'avs2',
    'b2b',
    'b2c',
    'b2e',
    'b8zs',
    'base64',
    'bit7',
    'bzip2',
    'bz2',
    'c2a',
    'c3p0',
    'c4s',
    'cast5',
    'ciks1',
    'coconut98',
    'code128',
    'code128a',
    'code128b',
    'code128c',
    'code39',
    'color16',
    'color32',
    'com1',
    'com2',
    'com3',
    'com4',
    'crc32',
    'cvv2',
    'ddr2',
    'ddr3',
    'ddr4',
    'd2d',
    'd2d2t',
    'd3d',
    'db2',
    'db9',
    'des3',
    'ds0',
    'ds1',
    'ean128',
    'ean13',
    'e2e',
    'e3cp',
    'ec2',
    'f2f',
    'f95',
    'fat32',
    'float128',
    'float32',
    'float64',
    'float96',
    'fma99',
    'fs1015',
    'fs1016',
    'fs1023',
    'g711',
    'g718',
    'g719',
    'g721',
    'g722',
    'g723',
    'g726',
    'g728',
    'g729',
    'g729a',
    'g729d',
    'gtk3',
    'h261',
    'h262',
    'h263',
    'h264',
    'h265',
    'h323',
    'hdf5',
    'hi10p',
    'hi422p',
    'hi444pp',
    'hl7',
    'html3',
    'html4',
    'html40',
    'html5',
    'i10n',
    'i18n',
    'i2c',
    'i386',
    'ia32',
    'ia64',
    'id10t',
    'ie10',
    'ie11',
    'ie12',
    'ie13',
    'ie6',
    'ie7',
    'ie8',
    'ie9',
    'imap4',
    'int128',
    'int16',
    'int2of5',
    'int32',
    'int64',
    'int8',
    'jpa2',
    'ipv4',
    'ipv6',
    'ipv6cp',
    'iso[0-9]{2,}',
    'ix86',
    'j2ee',
    'j2me',
    'j2se',
    'java10',
    'java6',
    'java7',
    'java8',
    'java9',
    'jdk10',
    'jdk11',
    'jdk12',
    'jdk13',
    'jdk14',
    'jdk15',
    'jdk16',
    'jpeg2000',
    'koi8r',
    'l10n',
    'l2f',
    'l2s',
    'l2tp',
    'l2tp',
    'l3f',
    'l3s',
    'l3tp',
    'log10',
    'log2',
    'log4j',
    'lpt1',
    'lpt2',
    'lpt4',
    'lpt4',
    'md5',
    'md5sum',
    'mp3',
    'mp4',
    'mpeg1',
    'mpeg2',
    'mpeg25',
    'mpeg3',
    'mpeg4',
    'multi2',
    'ns3',
    'nw7',
    'o2o',
    'oauth1',
    'oauth2',
    'os10',
    'os11',
    'os12',
    'os13',
    'os7',
    'os8',
    'os9',
    'p2p',
    'p2sc',
    'p3p',
    'p4m',
    'pep8',
    'perl5',
    'pop3',
    'px64',
    'python2',
    'python3',
    'qt4',
    'r2d2',
    'rc5',
    'rc6',
    'rfc[0-9]{3,}',
    'rj11',
    'rj45',
    'rot13',
    'scc14shippingcode',
    'sha0',
    'sha1',
    'sha1024',
    'sha1sum',
    'sha224sum',
    'sha224',
    'sha256sum'
    'sha256',
    'sha384sum',
    'sha384',
    'sha512sum',
    'sha512',
    'sint16',
    'sint32',
    'sint64',
    'sint8',
    'sm4',
    'smb2',
    'socks4',
    'ss7',
    'sscc18',
    'std2of5',
    'sys32',
    'uint16',
    'uint32',
    'uint64',
    'uint8',
    'usd3',
    'usd4',
    'utf16',
    'utf32',
    'utf8',
    'v2p',
    'vc1',
    'vc3',
    'vcard3',
    'vcard4',
    'vp3',
    'vp4',
    'vp5',
    'vp6',
    'vp6e',
    'vp6s',
    'vp7',
    'vp8',
    'vp9',
    'w3af',
    'w3c',
    'win32',
    'win64',
    'windows7',
    'windows10',
    'windows11',
    'x11',
    'x11r4',
    'x11r5',
    'x11r6',
    'x208',
    'x209',
    'x21',
    'x214',
    'x215',
    'x216',
    'x217',
    'x219',
    'x224',
    'x225',
    'x226',
    'x227',
    'x229',
    'x25',
    'x264',
    'x265',
    'x28',
    'x29',
    'x3d',
    'x3j16',
    'x3t10',
    'x400',
    'x409',
    'x500',
    'x509',
    'x64',
    'x680',
    'x75',
    'x86',
    'xfree86',
    'xga2',
    'xml10',
    'xml11',
    'y2k',
}
'''
Set of common abbreviations and symbols that contain numbers. Entries can be
regular expressions.
'''

special_computing_terms = {
    'adware',
    'ascii',
    'autoexec',
    'autosave',
    'autocommit',
    'backend',
    'backlink',
    'backprop',
    'backreference',
    'barcode',
    'bboard',
    'bitblt',
    'bitbucket',
    'bitcoin',
    'bitmap',
    'blocklist',
    'bnf',
    'btree',
    'builtin',
    'bytecode',
    'bzip',
    'callback',
    'camelcase',
    'charset',
    'chipset',
    'cgi',
    'checkbox',
    'classpath',
    'copyleft',
    'crosshair',
    'dataflow',
    'datastore',
    'deadlock',
    'defragment',
    'denoise',
    'deprecated',
    'devop',
    'devops',
    'distro',
    'dns',
    'dram',
    'ebnf',
    'ecmascript',
    'eeprom',
    'embeddable',
    'esata',
    'exec',
    'ext',
    'facebook',
    'fifo',
    'fn',
    'fs',
    'github',
    'gmail',
    'gnupg',
    'groupware',
    'gzip',
    'hardwire',
    'hardwired',
    'hashmap',
    'hashset',
    'hmac',
    'hostname',
    'hyperscale',
    'inode',
    'instagram',
    'intel',
    'intra',
    'icq',
    'io',
    'ipfs',
    'iscsi',
    'itunes',
    'javabean',
    'javadoc',
    'javascript',
    'jdbc',
    'jmeter',
    'json',
    'jvm',
    'linker',
    'listserv',
    'llvm',
    'logon',
    'macos',
    'malware',
    'memoization',
    'metamodel',
    'microcode',
    'microkernel',
    'middleware',
    'millis',
    'mmx',
    'mpegts',
    'multipass',
    'multitasking',
    'msata',
    'nan',
    'ndn',
    'newline',
    'nvme',
    'online',
    'opcode',
    'opendocument',
    'overclock',
    'overclocked',
    'parallelizable',
    'petaflop',
    'petaflops',
    'phish',
    'phishing',
    'pickling',
    'prefill',
    'popup',
    'pos',
    'pseudocode',
    'ptr',
    'quickbasic',
    'quicktime',
    'refactoring',
    'regex',
    'req',
    'rmi',
    'scrollbar',
    'segfault',
    'shebang',
    'signedness',
    'socks',
    'sourceforge',
    'sunos',
    'str',
    'symlink',
    'sysop',
    'tarball',
    'thread',
    'throwable',
    'todo',
    'trackpad',
    'txt',
    'usec',
    'underclock',
    'voip',
    'warez',
    'worklist',
    'xml',
    'xterm',
    'xwindows',
    'zlib',
    # 2021/08/06: add identifiers
    'seperated',
    'seperator',
    'publickey',
    'privatekey',
    'instantiator',
    'abilities',
    'wikidata',
    'micro',
    'clob',
    'zxing',
    'securities'
}
'''
Terms that are not usually found in dictionaries, and should be considered
single words rather than split.  These tend to be specialized computing
terms, including terms that contain one or more dictionary words inside
them; for example, "checkbox" would in normal English be considered two
words, but has come to be accepted as a common neologism in computing.
Additional entries include very common abbreviations such as "str" which
are not found in dictionaries.
'''

# The following list of prefixes and suffixes started as the lists from the
# web page for Samurai, https://hiper.cis.udel.edu/Samurai/Samurai.html
#
# Enslen, E., Hill, E., Pollock, L., & Vijay-Shanker, K. (2009).
# Mining source code to automatically split identifiers for software analysis.
# In Proceedings of the 6th IEEE International Working Conference on Mining
# Software Repositories (MSR'09) (pp. 71-80).
#
# I made additions to the prefixes.

prefixes = {'afro', 'ambi', 'amphi', 'ana', 'anglo', 'apo', 'astro', 'bi',
            'bio', 'circum', 'cis', 'co', 'col', 'com', 'con', 'contra',
            'cor', 'cryo', 'crypto', 'de', 'de', 'demi', 'di', 'dif',
            'dis', 'du', 'duo', 'eco', 'electro', 'em', 'en', 'epi',
            'euro', 'ex', 'franco', 'geo', 'hemi', 'hetero', 'homo',
            'hydro', 'hypo', 'ideo', 'idio', 'il', 'im', 'infra', 'inter',
            'intra', 'ir', 'iso', 'macr', 'mal', 'maxi', 'mega', 'megalo',
            'micro', 'midi', 'mini', 'mis', 'mon', 'multi', 'neo', 'omni',
            'paleo', 'para', 'ped', 'peri', 'poly', 'pre', 'preter',
            'proto', 'pyro', 're', 'retro', 'semi', 'socio', 'supra',
            'sur', 'sy', 'syl', 'sym', 'syn', 'tele', 'trans', 'tri',
            'twi', 'ultra', 'un', 'uni', 'non'}

suffixes = {'a', 'ac', 'acea', 'aceae', 'acean', 'aceous', 'ade', 'aemia',
            'agogue', 'aholic', 'al', 'ales', 'algia', 'amine', 'ana',
            'anae', 'ance', 'ancy', 'androus', 'andry', 'ane', 'ar',
            'archy', 'ard', 'aria', 'arian', 'arium', 'ary', 'ase',
            'athon', 'ation', 'ative', 'ator', 'atory', 'biont', 'biosis',
            'cade', 'caine', 'carp', 'carpic', 'carpous', 'cele', 'cene',
            'centric', 'cephalic', 'cephalous', 'cephaly', 'chory',
            'chrome', 'cide', 'clast', 'clinal', 'cline', 'coccus',
            'coel', 'coele', 'colous', 'cracy', 'crat', 'cratic',
            'cratical', 'cy', 'cyte', 'derm', 'derma', 'dermatous', 'dom',
            'drome', 'dromous', 'eae', 'ectomy', 'ed', 'ee', 'eer', 'ein',
            'eme', 'emia', 'en', 'ence', 'enchyma', 'ency', 'ene', 'ent',
            'eous', 'er', 'ergic', 'ergy', 'es', 'escence', 'escent',
            'ese', 'esque', 'ess', 'est', 'et', 'eth', 'etic', 'ette',
            'ey', 'facient', 'fer', 'ferous', 'fic', 'fication', 'fid',
            'florous', 'foliate', 'foliolate', 'fuge', 'ful', 'fy',
            'gamous', 'gamy', 'gen', 'genesis', 'genic', 'genous', 'geny',
            'gnathous', 'gon', 'gony', 'grapher', 'graphy', 'gyne',
            'gynous', 'gyny', 'ia', 'ial', 'ian', 'iana', 'iasis',
            'iatric', 'iatrics', 'iatry', 'ibility', 'ible', 'ic',
            'icide', 'ician', 'ick obsolete', 'ics', 'idae', 'ide', 'ie',
            'ify', 'ile', 'ina', 'inae', 'ine', 'ineae', 'ing', 'ini',
            'ious', 'isation', 'ise', 'ish', 'ism', 'ist', 'istic',
            'istical', 'istically', 'ite', 'itious', 'itis', 'ity', 'ium',
            'ive', 'ization', 'ize', 'kinesis', 'kins', 'latry', 'lepry',
            'ling', 'lite', 'lith', 'lithic', 'logue', 'logist', 'logy',
            'ly', 'lyse', 'lysis', 'lyte', 'lytic', 'lyze', 'mancy',
            'mania', 'meister', 'ment', 'merous', 'metry', 'mo', 'morph',
            'morphic', 'morphism', 'morphous', 'mycete', 'mycetes',
            'mycetidae', 'mycin', 'mycota', 'mycotina', 'ness', 'nik',
            'nomy', 'odon', 'odont', 'odontia', 'oholic', 'oic', 'oid',
            'oidea', 'oideae', 'ol', 'ole', 'oma', 'ome', 'ont', 'onym',
            'onymy', 'opia', 'opsida', 'opsis', 'opsy', 'orama', 'ory',
            'ose', 'osis', 'otic', 'otomy', 'ous', 'para', 'parous',
            'pathy', 'ped', 'pede', 'penia', 'phage', 'phagia', 'phagous',
            'phagy', 'phane', 'phasia', 'phil', 'phile', 'philia',
            'philiac', 'philic', 'philous', 'phobe', 'phobia', 'phobic',
            'phony', 'phore', 'phoresis', 'phorous', 'phrenia', 'phyll',
            'phyllous', 'phyceae', 'phycidae', 'phyta', 'phyte',
            'phytina', 'plasia', 'plasm', 'plast', 'plasty', 'plegia',
            'plex', 'ploid', 'pode', 'podous', 'poieses', 'poietic',
            'pter', 'rrhagia', 'rrhea', 'ric', 'ry', 's', 'scopy',
            'sepalous', 'sperm', 'sporous', 'st', 'stasis', 'stat',
            'ster', 'stome', 'stomy', 'taxy', 'th', 'therm', 'thermal',
            'thermic', 'thermy', 'thon', 'thymia', 'tion', 'tome', 'tomy',
            'tonia', 'trichous', 'trix', 'tron', 'trophic', 'tropism',
            'tropous', 'tropy', 'tude', 'ty', 'ular', 'ule', 'ure',
            'urgy', 'uria', 'uronic', 'urous', 'valent', 'virile',
            'vorous', 'xor', 'y', 'yl', 'yne', 'zoic', 'zoon', 'zygous',
            'zyme'}



#!/usr/bin/env python -O

version = {
    "__version__"     : '0.2.2',
    "__title__"       : 'CodeToolkit',
    "__description__" : 'A toolkit for processing source code',
    "__license__"     : 'GNU General Public License, version 3.0',
}

import os
from   os import path
from   setuptools import setup, find_packages
import sys

reqs = [
    'six',
    'numpy',
    'gensim',
    'nltk',
    'textdistance[extras]',
    'jpype1',
    'spacy'
]

# The following reads the variables without doing an "import spiral",
# because the latter will cause the python execution environment to fail if
# any dependencies are not already installed -- negating most of the reason
# we're using setup() in the first place.  This code avoids eval, for security.

# Finally, define our namesake.

setup(
    name                 = version['__title__'].lower(),
    description          = version['__description__'],
    long_description     = 'The toolkit provides methods for processing source code.',
    version              = version['__version__'],
    # url                  = version['__url__'],
    # author               = version['__author__'],
    # author_email         = version['__email__'],
    license              = version['__license__'],
    keywords             = "program-comprehension code-processing",
    packages             = find_packages(),
    package_dir          = {
                                'codetoolkit': 'codetoolkit',
                                'codetoolkit.spiral': 'codetoolkit/spiral',
                                'codetoolkit.posse': 'codetoolkit/posse',
                            },
    package_data         = {
                                'codetoolkit': [
                                    'predicates.txt',
                                    'verb.txt',
                                    'CodeAnalysis.jar'
                                ],
                                'codetoolkit.spiral': ['data/*'],
                                'codetoolkit.posse': ['corpus/*', 'dicts/*']
                            },
    include_package_data = True,
    install_requires     = reqs,
    platforms            = 'any',
    python_requires  = '>=3.6',
)



#!/usr/bin/env python
# -*- encoding: utf-8 -*-

import functools
import re
import requests

class CodeTokenizer:

    def __init__(self, searchcode_host="http://10.177.21.165:8080"):
        self.searchcode_api = searchcode_host + "/api/codesearch/?q={}&per_page=1"

    def split_camelcase(self, camel_case: str, to_lower=True):
        delimited_case = camel_case.replace('...', '-DOTDOTDOT-')
        delimited_case = delimited_case.split('.')[-1]
        delimited_case = delimited_case.replace('[]', ' []')
        delimited_case = delimited_case.replace('-DOTDOTDOT-', ' ...')
        delimited_case = re.sub(r'_', " ", delimited_case).strip()
        delimited_case = re.sub(
            r'([A-Za-z]| )([0-9]+)([A-Za-z]|$| )', r'\1 \2 \3', delimited_case)
        delimited_case = re.sub(r'([A-Z]+)([A-Z][a-z])', r'\1 \2', delimited_case)
        delimited_case = re.sub(r'([A-Z]+)', r' \1', delimited_case)
        delimited_case = re.sub(r'\s+', ' ', delimited_case)
        delimited_case = delimited_case.strip()

        if to_lower:
            delimited_case = delimited_case.lower()
        return delimited_case

    def word_count(self, word):
        url = self.searchcode_api.format(word)
        response = requests.get(url)
        count = response.json()['totalHits']
        print(f"count of '{word}': {count}")
        return count

    def handle_digit(self, words: list, min_hit=50) -> list:
        result = [words[0]]
        for current_word in words[1:]:
            previous_word = result[-1]
            if previous_word[-1].isdigit() or current_word[0].isdigit():
                potential = previous_word + current_word
                if self.word_count(potential) >= min_hit:
                    result.pop(-1)
                    current_word = f"{previous_word}{current_word}"
            result.append(current_word)
        return result

    def tokenize(self, identifier: str, min_hit=50):
        delimited_case = self.split_camelcase(identifier, to_lower=False)
        words = delimited_case.split()
        print("words after splitting camel case:", words)
        words = self.handle_digit(words, min_hit=min_hit)
        print("words after handling digits:", words)

        # TODO: handle other cases.
        # for example, split 'getinstance' into 'get' and 'instance'

        return words


if __name__ == "__main__":
    tokenizer = CodeTokenizer()
    tokenizer.tokenize("md5sum")
    tokenizer.tokenize("parseISO8601")
    tokenizer.tokenize("getPDFv2")


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

if __name__ == "__main__":
    from codetoolkit.code_pos import CodePOS

    codepos = CodePOS.get_inst()
    for word, tag in codepos("add item to index"):
        print(word, "|", tag)


#!/usr/bin/env python
# -*- encoding: utf-8 -*-

NOUN_STOPWORDS = {
    "Foo", "Bar", "foo", "bar", "instance", "object", "set", "map", "list", "type", "name", "value", "util", "utils"
}

VERB_STOPWORDS = {
    "get", "set", "new", "create",
}